#include "common.h"
#include <XUTILITY>
#include <sys/stat.h>
#include "gbio.h"
#include "translate.h"

typedef int (*PReadHeaderProc)(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo);
typedef int (*PReadDataProc)(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData);
typedef int (*PReadPaletteProc)(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette);
typedef void (*PDeleteHeaderProc)(void* HeaderInfo);

struct ImageCodecStruct
{
	PReadHeaderProc headerproc;
	PReadDataProc dataproc;
	PReadPaletteProc paletteproc;
	PDeleteHeaderProc deleteproc;
};

const ImageCodecStruct ImageCodecs[] = 
{
	{ReadJPGHeader, ReadJPGData, ReadJPGPalette, DeleteJPGHeaderInfo},
	{ReadPNGHeader, ReadPNGData, ReadPNGPalette, DeletePNGHeaderInfo},
	{ReadBMPHeader, ReadBMPData, ReadBMPPalette, DeleteBMPHeaderInfo},
	{ReadTGAHeader, ReadTGAData, ReadTGAPalette, DeleteTGAHeaderInfo},
	{ReadGIFHeader, ReadGIFData, ReadGIFPalette, DeleteGIFHeaderInfo}
};

const int CodecCount = sizeof (ImageCodecs)/sizeof(ImageCodecStruct);

DLL_FUNC int ReadImageHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	for(int i = 0; i < CodecCount; i++)
	{
		file_seek(file, 0, SEEK_SET);
		if (ImageCodecs[i].headerproc(file, pImage, HeaderInfo) == GBM_ERR_OK)
		{
			return GBM_ERR_OK;
		}
	}
	return -1;
}

DLL_FUNC int ReadImageData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	int codec_num = ((pImage->dwFlags & 0xff00) >> 8) -1;
	if (codec_num < 0)
		return GBM_ERR_BAD_ARG;
	if (ImageCodecs[codec_num].dataproc(file, pImage, HeaderInfo, pData) == GBM_ERR_OK)
	{
		return GBM_ERR_OK;
	}
	
	return GBM_ERR_NOT_SUPP;
}

DLL_FUNC int ReadImagePalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	int codec_num = ((pImage->dwFlags & 0xff00) >> 8) -1;
	if (codec_num < 0)
		return GBM_ERR_BAD_ARG;
	
	if (ImageCodecs[codec_num].paletteproc(file, pImage, HeaderInfo, pPalette) == GBM_ERR_OK)
	{
		return GBM_ERR_OK;
	}
	
	return GBM_ERR_NOT_SUPP;
}

HBITMAP _LoadImageFile(FILE_HANDLE file, ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadImageHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		BITMAPINFO bmInfo;
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &bmInfo.bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		if ((pImage->dwFlags & 0xff0000) == OPTIONS_FLIP)
			pBitmapHeader->biHeight = -(int)pImage->iHeight;
		else
			pBitmapHeader->biHeight = (int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc,&bmInfo,DIB_RGB_COLORS,(void**)&pBits,NULL,0);
		if (!Bitmap)
		{
			file_close(file);
			return NULL;
		}
		if (ReadImageData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		if ((pImage->dwFlags & 0xff) == FORMAT_RGB)
		{
			RGBtoBGR(pBits,pImage);
		}

		if ((pImage->dwFlags & 0xff) == FORMAT_RGBA)
		{
			RGBAtoBGRA(pBits, pImage);
		}
	}
	else
	{
		PALETTE pal;
		if (ReadImagePalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
		{
			file_close(file);
			return NULL;
		}

		int ColorCount = 1 << pImage->iBitPerPixel;
		BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
		if (!pbmInfo)
		{
			file_close(file);
			return NULL;
		}
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		if ((pImage->dwFlags & 0xff0000) == OPTIONS_FLIP)
			pBitmapHeader->biHeight = -(int)pImage->iHeight;
		else
			pBitmapHeader->biHeight = (int)pImage->iHeight;

		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;

		int cf = pImage->dwFlags & 0xff;
		if (cf == FORMAT_BGR)
		{
			InvertPalette(&pal);
		} 
		for (int i = 0; i < ColorCount; i++)
		{
			pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].b;
			pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
			pbmInfo->bmiColors[i].rgbRed = pal.colors[i].r;
			pbmInfo->bmiColors[i].rgbReserved = 0;
		}
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
		if (!Bitmap)
		{
			int err = GetLastError();
			free(pbmInfo);
			file_close(file);
			return NULL;
		}
		if (ReadImageData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		free(pbmInfo);
	}

	file_close(file);
	return Bitmap;
}

DLL_FUNC HBITMAP LoadImageFileA(char* filename, ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	
	return _LoadImageFile(file, pImage);
}

DLL_FUNC HBITMAP LoadImageFileW(wchar_t* filename, ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;

	return _LoadImageFile(file, pImage);
}

DLL_FUNC void DeleteImageHeaderInfo(ImageStruct* pImage, void* HeaderInfo)
{
	int codec_num = ((pImage->dwFlags & 0xff00) >> 8) -1;
	if (codec_num < 0)
		return ;
	ImageCodecs[codec_num].deleteproc(HeaderInfo);
}