#include "jpg.h"
#include "gbio.h"
#include <setjmp.h>
#include "jpeg/JPEGLIB.H"
#include "translate.h"

//*************************************************************

struct my_error_mgr 
{
	struct jpeg_error_mgr pub;	/* "public" fields */
	
	jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

/*
* Here's the routine that will replace the standard error_exit method:
*/

METHODDEF(void)
my_error_exit (j_common_ptr cinfo)
{
	/* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
	my_error_ptr myerr = (my_error_ptr) cinfo->err;
	
	/* Always display the message. */
	/* We could postpone this until after returning, if we chose. */
	(*cinfo->err->output_message) (cinfo);
	
	/* Return control to the setjmp point */
	longjmp(myerr->setjmp_buffer, 1);
}

struct JPG_PRIV
{
	jpeg_decompress_struct info;
	my_error_mgr err;
};



//*************************************************************

DLL_FUNC int ReadJPGHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	if (!file || !pImage || ! HeaderInfo)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	JPG_PRIV* priv = new JPG_PRIV;
	if (!priv)
	{
		return GBM_ERR_MEM;
	}
	*HeaderInfo = 0;

	jpeg_decompress_struct* cinfo = &priv->info;
	
	memset(cinfo,0,sizeof(jpeg_decompress_struct));
	/* We use our private extension JPEG error handler.
	* Note that this struct must live as long as the main JPEG parameter
	* struct, to avoid dangling-pointer problems.
	*/
	my_error_mgr* jerr = &priv->err;
			
	/* Step 1: allocate and initialize JPEG decompression object */
	
	/* We set up the normal JPEG error routines, then override error_exit. */
	cinfo->err = jpeg_std_error(&(jerr->pub));
	jerr->pub.error_exit = my_error_exit;
	/* Establish the setjmp return context for my_error_exit to use. */
	if (setjmp(jerr->setjmp_buffer)) 
	{
    /* If we get here, the JPEG code has signaled an error.
	* We need to clean up the JPEG object, close the input file, and return.
		*/
		jpeg_destroy_decompress(cinfo);
		delete priv;
		return GBM_ERR_JPG_JMP;
	}
	/* Now we can initialize the JPEG decompression object. */
	jpeg_create_decompress(cinfo);
	
	/* Step 2: specify data source (eg, a file) */
	
	jpeg_stdio_src(cinfo, file);
	
	/* Step 3: read file parameters with jpeg_read_header() */
	
	(void) jpeg_read_header(cinfo, TRUE);
	/* We can ignore the return value from jpeg_read_header since
	*   (a) suspension is not possible with the stdio data source, and
	*   (b) we passed TRUE to reject a tables-only JPEG file as an error.
	* See libjpeg.doc for more info.
	*/
	cinfo->mem->max_memory_to_use = 500000;

	pImage->iWidth = cinfo->image_width;
	pImage->iHeight = cinfo->image_height;
	pImage->iBitPerPixel = (cinfo->num_components << 3);
	pImage->dwFlags = 0;
	pImage->dwFlags = FORMAT_RGB | IMAGE_JPG | OPTIONS_FLIP;
	

	*HeaderInfo = priv;

	return GBM_ERR_OK;
}

DLL_FUNC int ReadJPGData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	if (!file || !pImage || !HeaderInfo || !pData)
	{
		return GBM_ERR_BAD_ARG;
	}

	JPG_PRIV* priv = (JPG_PRIV*)HeaderInfo;
	
	jpeg_decompress_struct* cinfo = &priv->info;
	/* More stuff */
	JSAMPARRAY buffer;		/* Output row buffer */
	int row_stride;		/* physical row width in output buffer */
	

	/* Step 4: set parameters for decompression */
	
	/* In this example, we don't need to change any of the defaults set by
	* jpeg_read_header(), so we do nothing here.
	*/
	
	/* Establish the setjmp return context for my_error_exit to use. */
	if (setjmp(priv->err.setjmp_buffer)) 
	{
    /* If we get here, the JPEG code has signaled an error.
	* We need to clean up the JPEG object, close the input file, and return.
		*/
		jpeg_destroy_decompress(cinfo);
		delete priv;
		return GBM_ERR_JPG_JMP;
	}

	/* Step 5: Start decompressor */
	
	(void) jpeg_start_decompress(cinfo);
	/* We can ignore the return value since suspension is not possible
	* with the stdio data source.
	*/
	
	/* We may need to do some setup of our own at this point before reading
	* the data.  After jpeg_start_decompress() we have the correct scaled
	* output image dimensions available, as well as the output colormap
	* if we asked for color quantization.
	* In this example, we need to make an output work buffer of the right size.
	*/ 
	/* JSAMPLEs per row in output buffer */
	row_stride = cinfo->output_width * cinfo->output_components;
	/* Make a one-row-high sample array that will go away when done with image */
	buffer = (*cinfo->mem->alloc_sarray)
		((j_common_ptr) cinfo, JPOOL_IMAGE, row_stride, 1);
	
	/* Step 6: while (scan lines remain to be read) */
	/*           jpeg_read_scanlines(...); */
	
	/* Here we use the library's state variable cinfo.output_scanline as the
	* loop counter, so that we don't have to keep track ourselves.
	*/
	int shift = (-row_stride & 0x3);
	unsigned char* pBits = (unsigned char*)pData;
	
	while (cinfo->output_scanline < cinfo->output_height) 
	{
    /* jpeg_read_scanlines expects an array of pointers to scanlines.
	* Here the array is only one element long, but you could ask for
	* more than one scanline at a time if that's more convenient.
		*/
		(void) jpeg_read_scanlines(cinfo, buffer, 1);
		/* Assume put_scanline_someplace wants a pointer and sample count. */
		//    put_scanline_someplace(buffer[0], row_stride);
		memcpy(pBits,buffer[0],row_stride);
		pBits+= shift+row_stride;
	}
	/* Step 7: Finish decompression */
	
	(void) jpeg_finish_decompress(cinfo);
	/* We can ignore the return value since suspension is not possible
	* with the stdio data source.
	*/
	
	/* Step 8: Release JPEG decompression object */
	
	/* This is an important step since it will release a good deal of memory. */
	jpeg_destroy_decompress(cinfo);
	
	/* After finish_decompress, we can close the input file.
	* Here we postpone it until after no more JPEG errors are possible,
	* so as to simplify the setjmp error logic above.  (Actually, I don't
	* think that jpeg_destroy can do an error exit, but why assume anything...)
	*/

	/* At this point you may want to check to see whether any corrupt-data
	* warnings occurred (test whether jerr.pub.num_warnings is nonzero).
	*/
	
	delete priv;
	return GBM_ERR_OK;
}

DLL_FUNC int ReadJPGPalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	if (!file || !pImage || !HeaderInfo || !pPalette)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	int p;
	for ( p = 0; p < 0x100; p++ )
		pPalette->colors[p].r = pPalette->colors[p].g = pPalette->colors[p].b = (byte) p;
	return GBM_ERR_OK;
}

HBITMAP _LoadJPGFile(FILE_HANDLE file, ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadJPGHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		BITMAPINFO bmInfo;
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &bmInfo.bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = -(int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc,&bmInfo,DIB_RGB_COLORS,(void**)&pBits,NULL,0);
		if (!Bitmap)
		{
			file_close(file);
			return NULL;
		}
		if (ReadJPGData(file, pImage, HeaderInfo,pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		RGBtoBGR((byte*)pBits,pImage);
	}
	else
	{
		PALETTE pal;
		if (ReadJPGPalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
		{
			file_close(file);
			return NULL;
		}

		unsigned char* pBits;
		int ColorCount = 1 << pImage->iBitPerPixel;
		BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
		if (!pbmInfo)
			return NULL;
		BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = -(int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		for (int i = 0; i < ColorCount; i++)
		{
			pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].b;
			pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
			pbmInfo->bmiColors[i].rgbRed = pal.colors[i].r;
			pbmInfo->bmiColors[i].rgbReserved = 0;
		}
		HDC dc = GetDC(0);
		Bitmap = CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
		if (!Bitmap)
		{
			free(pbmInfo);
			file_close(file);
			return NULL;
		}
		if (ReadJPGData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		free(pbmInfo);

	}

	return Bitmap;
}

DLL_FUNC HBITMAP LoadJPGFileA(char* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	return _LoadJPGFile(file, pImage);
}

DLL_FUNC HBITMAP LoadJPGFileW(wchar_t* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	return _LoadJPGFile(file, pImage);
}

DLL_FUNC void DeleteJPGHeaderInfo(void* HeaderInfo)
{
	delete (JPG_PRIV*)HeaderInfo;
}