#include "tga.h"
#include "gbio.h"
#include "translate.h"
//*************************************************************

#define	low_byte(w)	((byte)  (          (w)&0x00ff)    )
#define	high_byte(w)	((byte) (((unsigned)(w)&0xff00)>>8))
#define	make_word(a,b)	(((word)a) + (((word)b) << 8))

#define	TGA_NO_IMAGE	0		/* No image data included in file    */
#define	TGA_UNCOMP_CM	1		/* Uncompressed, Color-Mapped (VDA/D */
					/* and Targa M-8 images)             */
#define	TGA_UNCOMP_RGB	2		/* Uncompressed, RGB images (eg: ICB */
					/* Targa 16, 24 and 32)              */
#define	TGA_UNCOMP_BW	3		/* Uncompressed, B/W images (eg:     */
					/* Targa 8 and Targa M-8 images)     */
#define	TGA_RL_CM	9		/* Run-length, Color-Mapped (VDA/D   */
					/* and Targa M-8 images)             */
#define	TGA_RL_RGB	10		/* Run-length, RGB images (eg: ICB   */
					/* Targa 16, 24 and 32)              */
#define	TGA_RL_BW	11		/* Run-length, B/W images (eg: Targa */
					/* 8 and Targa M-8)                  */
#define	TGA_COMP_CM	32		/* Compressed Color-Mapped (VDA/D)   */
					/* data using Huffman, Delta, and    */
					/* run length encoding               */
#define	TGA_COMP_CM_4	33		/* Compressed Color-Mapped (VDA/D)   */
					/* data using Huffman, Delta, and    */
					/* run length encoding in 4 passes   */

#define	IDB_ATTRIBUTES	0x0f		/* How many attrib bits per pixel    */
					/* ie: 1 for T16, 8 for T32          */
#define	IDB_ORIGIN	0x20		/* Origin in top left corner bit     */
					/* else its in bottom left corner    */
#define	IDB_INTERLEAVE	0xc0		/* Interleave bits as defined below  */
#define	IDB_NON_INT	0x00		/* Non-Interlaced                    */
#define	IDB_2_WAY	0x40		/* 2 way (even/odd) interleaving     */
#define	IDB_4_WAY	0x80		/* 4 way interleaving (eg: AT&T PC)  */

struct TGA_HEADER
{
	byte n_chars_in_id;		/* Length of identification text     */
	byte color_map_present;		/* 0 means no, 1 yes                 */
	byte image_type;		/* Type of image file, one of TGA_   */
	byte color_map_start_low;	/* These 5 bytes are only valid if   */
	byte color_map_start_high;	/* color_map_present is 1. They      */
	byte color_map_length_low;	/* Specify the size of the color map */
	byte color_map_length_high;	/* and where it starts from          */
	byte color_map_entry_bits;	/* Bits per color map entry          */
					/* Typically 15, 16, 24 or 32        */
	byte x_origin_low;
	byte x_origin_high;
	byte y_origin_low;
	byte y_origin_high;
	byte width_low;
	byte width_high;
	byte height_low;
	byte height_high;
	byte bpp;			/* Typically 16, 24 or 32            */
	byte image_descriptor;		/* Split into IDB_ bits              */
};

#define	SIZEOF_TGA_HEADER 18

struct TGA_PRIV
{
	TGA_HEADER header;
};
//***********************************************************************


DLL_FUNC int ReadTGAHeader(FILE_HANDLE file, ImageStruct* pImage, void** HeaderInfo)
{
	if (!file || !pImage || ! HeaderInfo)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	TGA_PRIV *priv = new TGA_PRIV;
	if (!priv)
	{
		return GBM_ERR_MEM;
	}
	*HeaderInfo = NULL;

	if (file_read(file, (char *) &(priv->header), SIZEOF_TGA_HEADER) <= 0)
	{
		delete priv;
		return GBM_ERR_READ;
	}

	switch ( priv->header.image_type )
	{
		case TGA_RL_BW:
		case TGA_UNCOMP_BW:
			if ( priv->header.bpp != 8 )
			{
				delete priv;
				return GBM_ERR_TGA_BAD_BPP;
			}
			break;
		case TGA_RL_CM:
		case TGA_UNCOMP_CM:
			if ( priv->header.bpp != 8 )
			{
				delete priv;
				return GBM_ERR_TGA_BAD_BPP;
			}
			if ( priv->header.color_map_entry_bits != 24 )
			{
				delete priv;
				return GBM_ERR_TGA_BAD_PAL;
			}
			pImage->dwFlags = FORMAT_BGR | IMAGE_TGA;
			break;
		case TGA_RL_RGB:
		case TGA_UNCOMP_RGB:
			if ( priv->header.bpp != 16 &&
			     priv->header.bpp != 24 &&
			     priv->header.bpp != 32 )
			{
				delete priv;
				return GBM_ERR_TGA_BAD_BPP;
			}
			if (priv->header.bpp == 24 || priv->header.bpp == 16)
				pImage->dwFlags = FORMAT_BGR | IMAGE_TGA;
			if (priv->header.bpp == 32)
				pImage->dwFlags = FORMAT_BGRA | IMAGE_TGA;
			break;
		default:
			{
				delete priv;
				return GBM_ERR_TGA_BAD_TYPE;
			}
	}

	pImage->iWidth	= make_word(priv->header.width_low , priv->header.width_high );
	pImage->iHeight = make_word(priv->header.height_low, priv->header.height_high);

	if ( pImage->iWidth <= 0 || pImage->iHeight <= 0 )
		return GBM_ERR_BAD_SIZE;

	pImage->iBitPerPixel = ( priv->header.bpp == 8 ) ? 8 : 24;
	*HeaderInfo = priv;

	return GBM_ERR_OK;
}

DLL_FUNC int ReadTGAData(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, void* pData)
{
	if (!file || !pImage || !HeaderInfo || !pData)
	{
		return GBM_ERR_BAD_ARG;
	}

	TGA_PRIV *priv = (TGA_PRIV *) HeaderInfo;
	
	byte* data = (byte*)pData;

	byte *p;
	int i, stride;

	file_seek(file, (long) (SIZEOF_TGA_HEADER + priv->header.n_chars_in_id), SEEK_SET);

	if ( priv->header.color_map_present )
	{
		int color_map_length;
		int bpp_pal = priv->header.color_map_entry_bits;
			/* Valid values are 32, 24, 16 and sometimes 15 */

		if ( bpp_pal == 15 ) 
			bpp_pal = 16;

		color_map_length = make_word(priv->header.color_map_length_low, priv->header.color_map_length_high);
		file_seek(file, (long) ((color_map_length * bpp_pal) / 8L), SEEK_CUR);
	}

	stride = ((pImage->iWidth*pImage->iBitPerPixel + 31) >> 5) << 2;
	p = data;
	if ( (priv->header.image_descriptor & IDB_ORIGIN) != 0 )
	{
		p += (pImage->iHeight-1) * stride;
		stride = -stride;
	}

	switch ( priv->header.image_type )
	{
	case TGA_RL_CM:
	case TGA_RL_BW:
	case TGA_RL_RGB:
	{
		AHEAD *ahead;
		int x = 0, y = 0;
		if ( (ahead = create_ahead(file)) == NULL )
		{
			delete priv;
			return GBM_ERR_MEM;
		}
		switch ( priv->header.bpp )
		{

		case 8:
			while ( y < pImage->iHeight )
			{
				int i, cnt = read_ahead(ahead);
				if ( cnt == -1 )
				{
					delete priv;
					destroy_ahead(ahead);
					return GBM_ERR_READ;
				}
				if ( cnt & 0x80 )
				{
					byte v = (byte) read_ahead(ahead);
					for ( i = 0x80; i <= cnt; i++ )
					{
						data[x++] = v;
						if ( x == pImage->iWidth )
						{ 
							x = 0; y++; data += stride; 
						}
					}
				}
			else
				for ( i = 0; i <= cnt; i++ )
				{
					data[x++] = (byte) read_ahead(ahead);
					if ( x == pImage->iWidth )
					{
						x = 0; y++; data += stride; 
					}
				}
			}
		break;

		case 16:
			while ( y < pImage->iHeight )
			{
				int i, cnt = read_ahead(ahead);
				if ( cnt == -1 )
				{
					delete priv;
					destroy_ahead(ahead);
					return GBM_ERR_READ;
				}
				if ( cnt & 0x80 )
				{
					byte v1 = (byte) read_ahead(ahead);
					byte v2 = (byte) read_ahead(ahead);
					for ( i = 0x80; i <= cnt; i++ )
					{
						p[x++] = v1;
						p[x++] = v2;
						if ( x == pImage->iWidth*2 )
						{ 
							x = 0; y++; p += stride; 
						}
					}
				}
				else
				for ( i = 0; i <= cnt; i++ )
				{
					p[x++] = (byte) read_ahead(ahead);
					p[x++] = (byte) read_ahead(ahead);
					if ( x == pImage->iWidth*2 )
					{
						x = 0; y++; p += stride; 
					}
				}
			}

		break;

		case 24:
			while ( y < pImage->iHeight )
			{
				int i, cnt = read_ahead(ahead);
				if ( cnt == -1 )
				{
					delete priv;
					destroy_ahead(ahead);
					return GBM_ERR_READ;
				}
				if ( cnt & 0x80 )
				{
					byte v1 = (byte) read_ahead(ahead);
					byte v2 = (byte) read_ahead(ahead);
					byte v3 = (byte) read_ahead(ahead);
					for ( i = 0x80; i <= cnt; i++ )
					{
						p[x++] = v1;
						p[x++] = v2;
						p[x++] = v3;
						if ( x == pImage->iWidth*3 )
						{
							x = 0; y++; p += stride; 
						}
					}
				}
				else
				for ( i = 0; i <= cnt; i++ )
				{
					p[x++] = (byte) read_ahead(ahead);
					p[x++] = (byte) read_ahead(ahead);
					p[x++] = (byte) read_ahead(ahead);
					if ( x == pImage->iWidth*3 )
					{
						x = 0; y++; p += stride; 
					}
				}
			}
		break;


		case 32:
			while ( y < pImage->iHeight )
			{
				int i, cnt = read_ahead(ahead);
				if ( cnt == -1 )
				{
					destroy_ahead(ahead);
					delete priv;
					return GBM_ERR_READ;
				}
				if ( cnt & 0x80 )
				{
					byte v1 = (byte) read_ahead(ahead);
					byte v2 = (byte) read_ahead(ahead);
					byte v3 = (byte) read_ahead(ahead);
					byte v4 = (byte) read_ahead(ahead); 
					for ( i = 0x80; i <= cnt; i++ )
					{
						p[x++] = v1;
						p[x++] = v2;
						p[x++] = v3;
						p[x++] = v4;
						if ( x == pImage->iWidth*4 )
						{
							x = 0; y++; p += stride; 
						}
					}
				}
				else
					for ( i = 0; i <= cnt; i++ )
					{
						p[x++] = (byte) read_ahead(ahead);
						p[x++] = (byte) read_ahead(ahead);
						p[x++] = (byte) read_ahead(ahead);
						p[x++] = (byte) read_ahead(ahead); 
						if ( x == pImage->iWidth*4 )
						{
							x = 0; y++; p += stride; 
						}
					}
			}
		break;

		}
		destroy_ahead(ahead);
	}
	break;


	case TGA_UNCOMP_BW:
	case TGA_UNCOMP_CM:
	case TGA_UNCOMP_RGB:
		switch ( priv->header.bpp )
		{

		case 8:
			for ( i = 0; i < pImage->iHeight; i++, p += stride )
				if ( file_read(file, p, pImage->iWidth) != pImage->iWidth )
				{
					delete priv;
					return GBM_ERR_READ;
				}
		break;


		case 16:
			for ( i = 0; i < pImage->iHeight; i++, p += stride )
			{
				if ( file_read(file, p, pImage->iWidth * 2) != pImage->iWidth * 2 )
				{
					delete priv;
					return GBM_ERR_READ;
				}
			}
			break;


		case 24:
			for ( i = 0; i < pImage->iHeight; i++, p += stride )
				if ( file_read(file, p, pImage->iWidth * 3) != pImage->iWidth * 3 )
				{
					delete priv;
					return GBM_ERR_READ;
				}
				break;

		case 32:
			{
				for ( i = 0; i < pImage->iHeight; i++, p += stride )
				{
					if (file_read(file, p, pImage->iWidth*4) != pImage->iWidth*4)
					{
						delete priv;
						return GBM_ERR_READ;
					}
				}
				
			}
			break;

		}
	break;

	}
	delete priv;
	return GBM_ERR_OK;
}

DLL_FUNC int ReadTGAPalette(FILE_HANDLE file, ImageStruct* pImage, void* HeaderInfo, PALETTE* pPalette)
{
	if (!file || !pImage || !HeaderInfo || !pPalette)
	{
		return GBM_ERR_BAD_ARG;
	}
	
	TGA_PRIV *priv = (TGA_PRIV *) HeaderInfo;
	if (!priv)
	{
		return GBM_ERR_BAD_ARG;
	}

	switch ( priv->header.image_type )
	{

	case TGA_UNCOMP_BW:
	case TGA_RL_BW:
	{
		int i;
		for (  i = 0; i < 0x100; i++ )
		pPalette->colors[i].b =
		pPalette->colors[i].g =
		pPalette->colors[i].r = (byte) i;
	}
	break;

	case TGA_UNCOMP_CM:
	case TGA_RL_CM:
	{
		int color_map_start, color_map_length, i;

		file_seek(file, (long) (SIZEOF_TGA_HEADER+priv->header.n_chars_in_id), SEEK_SET);

		color_map_start  = make_word(priv->header.color_map_start_low , priv->header.color_map_start_high );
		color_map_length = make_word(priv->header.color_map_length_low, priv->header.color_map_length_high);

		memset(pPalette->colors, 0, 0x100 * sizeof(PALCOLOR));
		for ( i = color_map_start; i < color_map_start + color_map_length; i++ )
		{
			byte b[3];

			if ( file_read(file, (char *) b, 3) != 3 )
				return GBM_ERR_READ;
			pPalette->colors[i].r = b[0];
			pPalette->colors[i].g = b[1];
			pPalette->colors[i].b = b[2];
		}
	}
	break;
	}
	pPalette->colorcount = (1 << pImage->iBitPerPixel);
	return GBM_ERR_OK;
}

HBITMAP _LoadTGAFile(FILE_HANDLE file,ImageStruct* pImage)
{
	void* HeaderInfo;
	HBITMAP Bitmap;

	if (ReadTGAHeader(file,pImage,&HeaderInfo) != GBM_ERR_OK)
	{
		file_close(file);
		return NULL;
	}

	if (pImage->iBitPerPixel > 8)
	{
		BITMAPINFO bmInfo;
		unsigned char* pBits;
		BITMAPINFOHEADER* pBitmapHeader = &bmInfo.bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = (int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		HDC dc = GetDC(0);
		Bitmap=CreateDIBSection(dc,&bmInfo,DIB_RGB_COLORS,(void**)&pBits,NULL,0);
		if (!Bitmap)
		{
			file_close(file);
			return NULL;
		}
		if (ReadTGAData(file, pImage, HeaderInfo,pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
	}
	else
	{
		PALETTE pal;
		if (ReadTGAPalette(file,pImage,HeaderInfo,&pal) != GBM_ERR_OK)
		{
			file_close(file);
			return NULL;
		}

		unsigned char* pBits;
		int ColorCount = 1 << pImage->iBitPerPixel;
		BITMAPINFO* pbmInfo = (BITMAPINFO*)malloc(sizeof(BITMAPINFO) + (ColorCount - 1) * sizeof(RGBQUAD));
		if (!pbmInfo)
			return NULL;
		BITMAPINFOHEADER* pBitmapHeader = &pbmInfo->bmiHeader;
		pBitmapHeader->biBitCount = pImage->iBitPerPixel;
		pBitmapHeader->biClrImportant = 0;
		pBitmapHeader->biClrUsed = 0;
		pBitmapHeader->biCompression = BI_RGB;
		pBitmapHeader->biHeight = (int)pImage->iHeight;
		pBitmapHeader->biPlanes = 1;
		pBitmapHeader->biSize = sizeof(BITMAPINFOHEADER);
		pBitmapHeader->biSizeImage = 0;
		pBitmapHeader->biWidth = pImage->iWidth;
		pBitmapHeader->biXPelsPerMeter = 0;
		pBitmapHeader->biYPelsPerMeter = 0;
		for (int i = 0; i < ColorCount; i++)
		{
			pbmInfo->bmiColors[i].rgbBlue = pal.colors[i].r;
			pbmInfo->bmiColors[i].rgbGreen = pal.colors[i].g;
			pbmInfo->bmiColors[i].rgbRed = pal.colors[i].b;
			pbmInfo->bmiColors[i].rgbReserved = 0;
		}
		HDC dc = GetDC(0);
		Bitmap = CreateDIBSection(dc, pbmInfo, DIB_RGB_COLORS, (void**)&pBits, NULL, 0);
		if (!Bitmap)
		{
			free(pbmInfo);
			file_close(file);
			return NULL;
		}
		if (ReadTGAData(file, pImage, HeaderInfo, pBits) != GBM_ERR_OK)
		{
			file_close(file);
			DeleteObject(Bitmap);
			return NULL;
		}
		free(pbmInfo);
	}

	return Bitmap;
}

DLL_FUNC HBITMAP LoadTGAFileA(char* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openA(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;
	
	return _LoadTGAFile(file, pImage);
}

DLL_FUNC HBITMAP LoadTGAFileW(wchar_t* filename,ImageStruct* pImage)
{
	FILE_HANDLE file = file_openW(filename);
	if (file == (FILE_HANDLE)-1)
		return NULL;

	return _LoadTGAFile(file, pImage);
}

DLL_FUNC void DeleteTGAHeaderInfo(void* HeaderInfo)
{
	delete (TGA_PRIV*)HeaderInfo;
}